#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: helpers.py
#
# Copyright 2019 Vincent Schouten
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
Import all parts from helpers here.

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html
"""
import json
import logging
import logging.config
import threading
import subprocess
import os.path
from urllib.error import URLError
import urllib.request
from socket import timeout
from time import sleep
from voluptuous import Schema, Required, Any, REMOVE_EXTRA, Optional, MultipleInvalid
from minitorcorelib import Tunnel, Machine  # , start_ping
from ..minitorcliexceptions import SetupFailed

__author__ = '''Vincent Schouten <inquiry@intoreflection.co>'''
__docformat__ = '''google'''
__date__ = '''06-04-2019'''
__copyright__ = '''Copyright 2019, Vincent Schouten'''
__license__ = '''MIT'''
__maintainer__ = '''Vincent Schouten'''
__email__ = '''<inquiry@intoreflection.co>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".


# This is the main prefix used for logging.
logging.basicConfig(format='%(asctime)s %(name)s %(levelname)s %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S')
LOGGER_BASENAME = '''helpers'''
LOGGER = logging.getLogger(LOGGER_BASENAME)  # non-class objects like functions can consult this Logger object
# LOGGER.addHandler(logging.NullHandler())  # method not in https://docs.python.org/3/library/logging.html

# Constants
HEARTBEAT_DURATION = 5

MODE_SCHEMA = Schema({Required("mode"): Any("TOR", "FOR", "COMMAND", "FILE"),
                      }, extra=REMOVE_EXTRA)

TOR_SCHEMA = Schema({"mode": "TOR",
                     "proxies": [{"ip_in": str,
                                  "ip_out": str,
                                  "identity_file": str,
                                  "hostname": str}],
                     "destination": {"ip_in": str,
                                     "ip_out": str,
                                     "identity_file": str,
                                     "hostname": str},
                     Optional("application"): {"binary_name": str,
                                               "binary_location": str}
                     }, required=True)

FOR_SCHEMA = Schema({"mode": "FOR",
                     "proxies": [{"ip_in": str,
                                  "ip_out": str,
                                  "identity_file": str,
                                  "hostname": str}],
                     "destination": {"ip_in": str,
                                     "identity_file": str,
                                     "hostname": str},
                     "forwarders": [{"local_port": int,
                                     "remote_interface": str,
                                     "remote_port": int}],
                     Optional("application"): {"binary_name": str,
                                               "binary_location": str}
                     }, required=True)

COMMAND_SCHEMA = Schema({"mode": "COMMAND",
                         "proxies": [{"ip_in": str,
                                      "ip_out": str,
                                      "identity_file": str,
                                      "hostname": str}],
                         "destination": {"ip_in": str,
                                         "identity_file": str,
                                         "hostname": str}
                         }, required=True)

FILE_SCHEMA = Schema({"mode": "FILE",
                      "proxies": [{"ip_in": str,
                                   "ip_out": str,
                                   "identity_file": str,
                                   "hostname": str}],
                      "destination": {"ip_in": str,
                                      "identity_file": str,
                                      "hostname": str},
                      "files": [{"source": str,
                                 "destination": str}]
                      }, required=True)


class Configuration:  # pylint: disable=too-few-public-methods, too-many-instance-attributes
    """Parses the configuration file."""

    def __init__(self, config_file):
        """Initializes the Configuration object.

        Args:
            config_file(basestring): A file containing a JSON document.

        """
        logger_name = u'{base}.{suffix}'.format(base=LOGGER_BASENAME,
                                                suffix='Configuration')
        self._logger = logging.getLogger(logger_name)
        try:
            config = self.get_config(config_file)
            self.mode = config.get('mode')
            self.proxies = config.get('proxies')
            self.application = config.get('application')
            self.destination = config.get('destination')
            self.files = config.get('files', '')
            forwarders = config.get('forwarders')
            self.forwarders_string = ' '.join([f'-L:{forwarder["local_port"]}:'
                                               f'{forwarder["remote_interface"]}:'
                                               f'{forwarder["remote_port"]}' for forwarder in
                                               forwarders]) if forwarders else ''
            self.forwarders_ports = ', '.join(
                [str(forwarder["local_port"]) for forwarder in forwarders]) if forwarders else ''
            self.destination['ip_out'] = self.destination.get('ip_out') or self.destination.get('ip_in')

        except AttributeError as exp:
            self._logger.error('configuration file could not be parsed. %s', exp)  # raised by config.get()
            raise SystemExit(1)

    def get_config(self, filename):
        """Validates the data structure and parses the parameters in a dictionary."""
        try:
            with open(filename, 'r') as file:
                _temp = json.loads(file.read())
                _mode = MODE_SCHEMA(_temp)
                if _mode.get('mode') == 'TOR':
                    self._logger.info('mode TOR enabled')
                    schema = TOR_SCHEMA(_temp)
                elif _mode.get('mode') == 'FOR':
                    self._logger.info('mode FOR enabled')
                    schema = FOR_SCHEMA(_temp)
                elif _mode.get('mode') == 'COMMAND':
                    self._logger.info('mode COMMAND enabled')
                    schema = COMMAND_SCHEMA(_temp)
                elif _mode.get('mode') == 'FILE':
                    self._logger.info('mode FILE enabled')
                    schema = FILE_SCHEMA(_temp)
                else:
                    self._logger.error('no mode enabled')
                    schema = None
                return schema
        except FileNotFoundError:  # open()
            self._logger.error('configuration file could not be opened ("FileNotFoundError")')
            return False
        except IOError:  # open()
            self._logger.error('not enough permissions to open configuration file ("IOError")')
        except ValueError:  # json.loads()
            self._logger.error('JSON document could not be deserialized ("ValueError")')
        except MultipleInvalid:  # Schema()
            self._logger.error('data structure (dict) validating failed ("MultipleInvalid")')


class Heartbeat:  # context manager
    """Determines periodically the state of the tunnel."""

    def __init__(self, local_heartbeat_port):
        logger_name = u'{base}.{suffix}'.format(base=LOGGER_BASENAME,
                                                suffix='Heartbeat')
        self._logger = logging.getLogger(logger_name)
        self.thread = None
        self.is_tunnel_intact = None
        self.terminate = False
        self.local_heartbeat_port = local_heartbeat_port

    def _run_heartbeat(self):
        while self.thread.is_alive:
            if self.terminate:
                return None
            self.is_tunnel_intact = start_ping(self.local_heartbeat_port)
            if self.is_tunnel_intact:
                self._logger.debug('heartbeat package was sent')
            sleep(HEARTBEAT_DURATION)

    def __enter__(self):
        self.thread = threading.Thread(target=self._run_heartbeat)
        self.thread.start()
        self._logger.info('heartbeat mechanism started')
        return self

    def __exit__(self, *exc):
        self.terminate = True
        self._logger.info('heartbeat mechanism stopped')


class StateManager:  # context manager
    """Cleans up active Tunnel(s) and/or Machine(s) when exiting.

    Returns:
        self (StateManager): The instantiated StateManager object.

    """

    def __init__(self):
        logger_name = u'{base}.{suffix}'.format(base=LOGGER_BASENAME,
                                                suffix='StateManager')
        self._logger = logging.getLogger(logger_name)
        self._running_instances = []

    def __enter__(self):
        self._logger.info('opening Tunnels and starting Machines...')
        return self

    def add_object(self, object_name):
        """Collects instantiated Tunnel(s) and instantiated Machine(s) for clean up purposes."""
        self._running_instances.append(object_name)

    def _clean_up(self):
        for instance in reversed(self._running_instances):
            if isinstance(instance, Tunnel):
                if instance.close():
                    self._logger.debug('close() method invoked on <%s>', instance.__class__.__name__)
                else:
                    self._logger.error('close() method could not be invoked on <%s>', instance.__class__.__name__)
            elif isinstance(instance, Machine):
                if instance.stop():
                    self._logger.debug('stop() method invoked on <%s>', instance.__class__.__name__)
                else:
                    self._logger.error('stop() method could not be invoked on <%s>', instance.__class__.__name__)

    def __exit__(self, *exc):
        self._logger.info('closing Tunnels and stopping Machines...')
        self._clean_up()


def setup_link(state, transfer_agent, tunnel, machine):
    """Starts instantiated TransferAgents, opens instantiated Tunnel(s) and starts instantiated Machine(s).

    Args:
        state (StateManager): An instantiated StateManager object.
        transfer_agent (TransferAgent): An instantiated TransferAgent object.
        tunnel (Tunnel): An instantiated Tunnel object.
        machine (Machine): An instantiated Machine object.

    """
    if not transfer_agent.start():
        raise SetupFailed(transfer_agent)
    if transfer_agent.__class__.__name__ == 'FirstTransferAgent':
        LOGGER.info('Agent transferred to %s ', machine.hostname)
    elif transfer_agent.__class__.__name__ == 'ConsecutiveTransferAgent':
        LOGGER.info('Agent transferred through Tunnel to %s ', machine.hostname)

    state.add_object(tunnel)
    if not tunnel.open():
        raise SetupFailed(tunnel)
    LOGGER.info('Tunnel opened (%s)',
                tunnel.ip_address_i)

    state.add_object(machine)
    if not machine.start():
        raise SetupFailed(machine)
    LOGGER.info('Machine started (%s)',
                machine.hostname)


def start_application(binary_name, binary_location):  # used in either FOR (w/ Thunderbird) or TOR mode (w/ Firefox)
    """Starts the application."""
    try:
        # process = subprocess.run([binary_location], capture_output=False)
        # process = subprocess.run([binary_location], subprocess.DEVNULL)
        with open(os.devnull, 'w') as devnull:
            process = subprocess.run([binary_location], stdout=devnull, stderr=devnull)
        LOGGER.debug('application %s executed and the return code was: %s', binary_name, process.returncode)
        return True
    except FileNotFoundError:
        LOGGER.error('application (rather executable binary) of %s was not found', binary_name)
        return False


def start_ping(local_heartbeat_port):  # TIMEOUT needed?
    """Sends a HTTP GET request and processes the response."""
    local_heartbeat_port = local_heartbeat_port
    http_code = 0
    result = False
    try:
        with urllib.request.urlopen(f'http://localhost:{local_heartbeat_port}', timeout=5, data=None) as request_obj:
            http_code = request_obj.getcode()
    except URLError:
        LOGGER.error('minitoragent did not respond to GET request. '
                     'probable cause: Machine unreachable or client has no connection to the Internet.')
        result = False
    except ConnectionResetError:
        LOGGER.error('minitoragent did not respond to GET request. '
                     'probable cause: minitoragent not bind to port.')
        result = False
    except timeout:
        LOGGER.error('minitoragent did not respond to GET request in a timely fashion. '
                     'probable cause: Machine unreachable or client has no connection to the Internet.')
        result = False
    if http_code == 200:
        result = True
    return result
