#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: minitorcli.py
#
# Copyright 2019 Vincent Schouten
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
Main code for minitorcli.

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html

"""

# TODO:  pylint: disable=fixme
#  modify JSON document to include:
#     - the MACHINE_DEPLOY_PATH (default: /tmp/)
#     - the LOCAL_HEARTBEAT_PORT (default: 11600)
#     - the LOCAL_PROXY_PORT (default: 8080)
#     - the MACHINE_PORT_AGENT (default: 44191)
#     - the MACHINE_PORT_PROXY_SERVER (default: 44192)
#     - the MACHINE_PORT_HEARTBEAT_RESPONDER (default: 44193)


import argparse
import logging.config
from time import sleep

import coloredlogs
from minitorcorelib import (PortGenerator,
                            TransferAgentFactory,
                            MachineFactory,
                            TunnelFactory)

from .lib import Heartbeat, StateManager, Configuration, setup_link, start_application
from .minitorcliexceptions import SetupFailed

__author__ = '''Vincent Schouten <inquiry@intoreflection.co>'''
__docformat__ = '''google'''
__date__ = '''06-04-2019'''
__copyright__ = '''Copyright 2019, Vincent Schouten'''
__credits__ = ["Vincent Schouten"]
__license__ = '''MIT'''
__maintainer__ = '''Vincent Schouten'''
__email__ = '''<inquiry@intoreflection.co>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".

# This is the main prefix used for logging
LOGGER_BASENAME = '''minitorcli'''  # non-class objects like functions can consult this Logger object
LOGGER = logging.getLogger(LOGGER_BASENAME)
# LOGGER.addHandler(logging.NullHandler())  # method not in https://docs.python.org/3/library/logging.html

# Constants
LOCAL_HEARTBEAT_PORT = 11600  # a local port used by the heartbeat mechanism to communicate with the last Machine
LOCAL_PROXY_PORT = 8080  # a local port used to forward web traffic which exits last Machine (only in TOR mode)
LOCAL_TRANSFER_PORT = 11700  # a local port used by minitor to upload files to last Machine (only in FILE mode)
LOCAL_COMMAND_PORT = 11800  # a local port used by minitor to send linux commands to last Machine (only in COMMAND mode)
MACHINE_DEPLOY_PATH = '/tmp/'  # the path on Machine where the Agent will be transferred to


def get_arguments():
    """
    Gets us the cli arguments.

    Returns the args as parsed from the argsparser.
    """
    # https://docs.python.org/3/library/argparse.html
    parser = argparse.ArgumentParser(
        description='''minitor - anonymizing internet traffic using private servers (cli)''')
    parser.add_argument('--config-file',
                        '-c',
                        help='The location of the config file',
                        dest='config_file',
                        action='store',
                        default='')
    parser.add_argument('--log-level',
                        '-L',
                        help='Provide the log level. Defaults to info.',
                        dest='log_level',
                        action='store',
                        default='info',
                        choices=['debug',
                                 'info',
                                 'warning',
                                 'error',
                                 'critical'])
    args = parser.parse_args()
    return args


def main():  # pylint: disable=too-many-statements, too-many-locals
    """
    Main method.

    This method holds what you want to execute when
    the script is run on command line.
    """
    args = get_arguments()
    coloredlogs_format = '%(asctime)s %(name)s[%(process)d] %(levelname)s %(message)s'
    coloredlogs.install(fmt=coloredlogs_format, level=args.log_level.upper())
    config = Configuration(args.config_file)
    port_generator = PortGenerator()
    try:
        with StateManager() as state:
            for proxy in config.proxies:
                ports = port_generator.get_next_port_ranges()
                transfer_agent = TransferAgentFactory(local_tunnel_fw_port=ports['local_tunnel_fw_port'],
                                                      ip_address_i=proxy['ip_in'],
                                                      identity_file=proxy['identity_file'],
                                                      machine_deploy_path=MACHINE_DEPLOY_PATH)
                tunnel = TunnelFactory(local_tunnel_fw_port=ports['local_tunnel_fw_port'],
                                       local_agent_port=ports['local_agent_port'],
                                       count_connection=ports['count_connection'],
                                       ip_address_i=proxy['ip_in'],
                                       identity_file=proxy['identity_file'])
                machine = MachineFactory(tunnel=tunnel,
                                         hostname=proxy['hostname'],
                                         machine_deploy_path=MACHINE_DEPLOY_PATH,
                                         ip_address_e=proxy['ip_out'])
                setup_link(state, transfer_agent, tunnel, machine)
            ports = port_generator.get_next_port_ranges()
            transfer_agent = TransferAgentFactory(local_tunnel_fw_port=ports['local_tunnel_fw_port'],
                                                  ip_address_i=config.destination['ip_in'],
                                                  identity_file=config.destination['identity_file'],
                                                  machine_deploy_path=MACHINE_DEPLOY_PATH)
            tunnel = TunnelFactory(local_tunnel_fw_port=ports['local_tunnel_fw_port'],
                                   local_agent_port=ports['local_agent_port'],
                                   count_connection=ports['count_connection'],
                                   ip_address_i=config.destination['ip_in'],
                                   identity_file=config.destination['identity_file'],
                                   mode=config.mode,
                                   local_heartbeat_port=LOCAL_HEARTBEAT_PORT,
                                   local_browser_port=LOCAL_PROXY_PORT,  # exclusively used in TOR mode
                                   local_transfer_port=LOCAL_TRANSFER_PORT,  # exclusively used in FILE mode
                                   local_command_port=LOCAL_COMMAND_PORT,  # exclusively used in COMMAND mode
                                   local_forward_connections=config.forwarders_string)  # exclusively used in FOR mode
            machine = MachineFactory(tunnel=tunnel,
                                     hostname=config.destination['hostname'],
                                     machine_deploy_path=MACHINE_DEPLOY_PATH,
                                     mode=config.mode,
                                     ip_address_e=config.destination['ip_out'],  # exclusively used in TOR mode
                                     transfer_port=LOCAL_TRANSFER_PORT,  # exclusively used in FILE mode
                                     command_port=LOCAL_COMMAND_PORT)  # exclusively used in COMMAND mode
            setup_link(state, transfer_agent, tunnel, machine)
            if config.mode == 'FOR':
                LOGGER.info('connections on local ports %s will be forwarded', config.forwarders_ports)
            elif config.mode == 'TOR':
                LOGGER.info('local port %s will be listening for web traffic', LOCAL_PROXY_PORT)
            elif config.mode == 'COMMAND':
                shell_machine = machine
            elif config.mode == 'FILE':
                file_machine = machine
            else:
                raise SystemExit(1)
            with Heartbeat(LOCAL_HEARTBEAT_PORT) as heartbeat:
                LOGGER.info('encrypted tunnel established')
                sleep(0.4)
                if config.application:  # minitorcli will *not* terminate tunnel when no heartbeats were received
                    LOGGER.info('starting application...')
                    start_application(binary_name=config.application['binary_name'],
                                      binary_location=config.application['binary_location'])
                elif config.mode == 'COMMAND':  # minitorcli 'll *not* terminate tunnel when no heartbeats were received
                    LOGGER.warning('the interface does not support shell meta characters ')
                    LOGGER.warning('such as pipe and it\'s not possible to interact with ')
                    LOGGER.warning('programs that need a response. hit control-c to quit')
                    try:
                        while True:
                            command = input('enter command: ')
                            response_raw = shell_machine.exec_command(command)  # type of response is 'byte'
                            response_str = response_raw.decode("utf-8")
                            response_line = response_str.split('\n')
                            for line in response_line:
                                print('>    %s' % line)
                    except KeyboardInterrupt:
                        raise SystemExit(0)
                elif config.mode == 'FILE':
                    file_machine.transfer(metadata_files=config.files)
                else:  # minitorcli *will* terminate tunnel when no heartbeats were received
                    while True:
                        if not heartbeat.is_tunnel_intact:  # in next release: only break tunnel -
                            LOGGER.error('encrypted tunnel has been broken')  # when there was no heartbeat for t
                            raise KeyboardInterrupt
                        else:
                            sleep(1)
    except KeyboardInterrupt:
        LOGGER.info('encrypted tunnel dismantled')
        raise SystemExit(0)
    except SetupFailed as msg:  # defined in "minitorcliexceptions" and can only be raised by setup_link() in helpers
        LOGGER.error(msg)
        raise SystemExit(1)
